#!/usr/bin/perl -w

###########################################################################
#
# inittranslation.pl
#
# A component of the Greenstone digital library software
# from the New Zealand Digital Library Project at the 
# University of Waikato, New Zealand.
#
# Copyright (C) 1999 New Zealand Digital Library Project
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
###########################################################################


BEGIN {
    die "GSDLHOME not set\n" unless defined $ENV{'GSDLHOME'};
    unshift (@INC, "$ENV{'GSDLHOME'}/perllib");
}


###########################################################################
#
# CODE SYNOPSIS
#
# This script initialises a translation using the Translator Interface. It:
#
# - Parses the script arguments.
# - Checks that the source and target languages are specified in the
#     etc/main.cfg file.
# - Reads the translation files from the translator configuration file,
#     and check that the one specified is defined.
# - Checks that the translation hasn't previously been initialised.
# - Creates a new directory for the translation databases and files.
# - Reads the translation file data from the config file.
# - Copies the source language file into the translation directory.
# - Parses the (copied) source language file.
# - Creates a new translation database.
# - Parses the target language file.
# - Adds the current contents of the source and target language files
#     to the translation database.
# - Uses CVS to mark the translations needing updating.
#
###########################################################################


use translator;
use util;
use GDBM_File;


my $gsdldir = "$ENV{'GSDLHOME'}";
my $cvsroot = &translator::get_cvs_root();

my $option_not_in_cvs = 0;


sub main
{
    # Get the code of the source (base) language
    local $sourceabbr = shift(@_);
    # Get the code of the target language
    local $targetabbr = shift(@_);
    # Get the name of the file to translate
    local $transfile = shift(@_);

    # Read any extra arguments
    local $optionalarg = shift(@_);
    while ($optionalarg) {
	if ($optionalarg eq "-not_in_cvs") { $option_not_in_cvs = 1; }
	$optionalarg = shift(@_);
    }

    # Log any configuration errors to the general translator log
    local $translatordir = &translator::get_translator_dir();
    local $translatorlogfile = &util::filename_cat($translatordir, "translator.log");
    if (!open(LOG, ">>$translatorlogfile")) {
	die "Error: Could not append to log file $translatorlogfile.\n";
    }

    # Log the command that launched this script
    print LOG "\nCommand: $0 @ARGV\n";

    # Check that the necessary arguments were supplied
    if (!$sourceabbr || !$targetabbr || !$transfile) {
	print LOG &get_usage_description();
	print LOG "Error: Required argument missing.\n";
	die "\n";
    }

    # Check that the languages are specified in the etc/main.cfg file
    local %langhash = &translator::read_languages_from_main_cfg_file();
    if (!$langhash{$sourceabbr}) {
	print LOG "Error: Language \"$sourceabbr\" is not declared in etc/main.cfg.\n";
	die "\n";
    }
    if (!$langhash{$targetabbr}) {
	print LOG "Error: Language \"$targetabbr\" is not declared in etc/main.cfg.\n";
	die "\n";
    }

    # Check that the file to translate is defined in the translator config file
    local %files = &translator::read_translation_files_from_config_file();
    if (!$files{$transfile}) {
	print LOG "Error: File \"$transfile\" is not defined in etc/translator.cfg.\n";
	die "\n";
    }

    # Make sure a directory exists for translations between the source and target languages
    local $translatordir = &translator::get_translator_dir();
    local $languagedir = &util::filename_cat($translatordir, "$sourceabbr-$targetabbr");
    if (!-e $languagedir) {
	if (!mkdir($languagedir, 0777)) {
	    print LOG "Error: Couldn't create directory $languagedir.\n";
	    die "\n";
	}
    }

    # Make sure a translation process for this file hasn't already been started
    local $translationdir = &util::filename_cat($languagedir, $transfile);
    local $translationdb = &util::filename_cat($translationdir, "translation.db");
    if (-e $translationdb) {
	print LOG "Note: Translation database already exists for $transfile.\n";
	return;
    }

    # Create a directory to store the translation data for this file
    &util::mk_all_dir($translationdir);

    # Read the details of the file to translate from the translator config file
    local ($sourcefile, $targetfile, $filetype) =
	&translator::read_file_data_from_config_file($sourceabbr, $targetabbr, $transfile);

    # Log all the rest of the output to the log for this translation process
    close(LOG);
    local $translationlogfile = &util::filename_cat($translationdir, "translation.log");
    if (!open(LOG, ">>$translationlogfile")) {
	die "Error: Could not append to log file $translationlogfile.\n";
    }

    # Log the command that launched this script
    print LOG "\nCommand: $0 @ARGV\n";


    # -----------------------------------------
    #   Process the source language file
    # -----------------------------------------

    # If the source language is not English, the source files may not be up to date
    if ($sourceabbr !~ /en/) {
	print LOG "Warning: Source language is not English. You should check that" .
	          " the source language files are up to date before beginning.\n";
    }

    # Check that the source file exists
    local $hotsourcepath = &util::filename_cat($gsdldir, $sourcefile);
    if (! -e $hotsourcepath) {
	print LOG "Error: Source file $hotsourcepath does not exist.\n";
	die "\n";
    }

    # Check that the source file is up to date
    unless ($option_not_in_cvs == 1) {
	local $sourceupdate = `cd $gsdldir; cvs -d $cvsroot update $sourcefile`;
	if ($sourceupdate =~ /^C /) {
	    print LOG "Error: Source file conflicts with the repository.\n";
	    die "\n";
	}
	if ($sourceupdate =~ /^M /) {
	    print LOG "Error: Source file contains uncommitted changes.\n";
	    die "\n";
	}
    }

    # Make a backup of the source file in the translation directory
    local $sourcepath = &translator::get_translation_file_path($translationdir, $sourcefile);
    &util::cp($hotsourcepath, $sourcepath);

    # Parse the source language file, building a mapping from chunk key to chunk text
    local %sourcelinehash = &translator::build_key_to_line_mapping($sourcepath, $filetype);
    local %sourcehash = &translator::build_key_to_text_mapping($sourcepath, $filetype,
							       %sourcelinehash);
    print LOG "Number of source chunks: " . scalar(keys(%sourcehash)) . "\n";


    # -----------------------------------------
    #   Initialise the translation database
    # -----------------------------------------

    # Create a database to store the translation information for this file
    local %translationdata = ();
    tie(%translationdata, "GDBM_File", $translationdb, GDBM_WRCREAT, 0666);

    # Store a mapping from key to target text (initially blank)
    foreach $key (keys(%sourcehash)) {
	# print "$key:";
	# print "$sourcehash{$key}\n";
	$translationdata{$key} = "";
    }

    # We've finished initialising the translation database
    untie(%translationdata);


    # -----------------------------------------
    #   Process the target language file
    # -----------------------------------------

    # If the target file doesn't exist, we're done
    local $hottargetpath = &util::filename_cat($gsdldir, $targetfile);
    if (! -e $hottargetpath) {
	print LOG "Note: Target file $hottargetpath does not exist.\n";
	return;
    }

    # Check that the target file is up to date
    unless ($option_not_in_cvs == 1) {
	local $targetupdate = `cd $gsdldir; cvs -d $cvsroot update $targetfile`;
	if ($targetupdate =~ /^C /) {
	    print LOG "Error: Target file conflicts with the repository.\n";
	    die "\n";
	}
	if ($targetupdate =~ /^M /) {
	    print LOG "Error: Target file contains uncommitted changes.\n";
	    die "\n";
	}
    }

    # Make a backup of the target file in the translation directory
    local $targetpath = &translator::get_translation_file_path($translationdir, $targetfile);
    &util::cp($hottargetpath, $targetpath);
    print STDERR "Target path: $targetpath\n";

    # Parse the target language file, building a mapping from chunk key to chunk text
    local %targetlinehash = &translator::build_key_to_line_mapping($targetpath, $filetype);
    local %targethash = &translator::build_key_to_text_mapping($targetpath, $filetype,
							       %targetlinehash);
    print LOG "Number of target chunks: " . scalar(keys(%targethash)) . "\n";

    # -----------------------------------------
    #   Populate the translation database
    # -----------------------------------------

    # Store a mapping from key to target text
    tie(%translationdata, "GDBM_File", $translationdb, GDBM_WRCREAT, 0666);
    foreach $key (keys(%sourcehash)) {
	if ($targethash{$key}) {
	    $translationdata{$key} = $targethash{$key};
	    # print "Mapping $key -> " . $translationdata{$key} . "\n";
	}
    }
    untie(%translationdata);

    # -----------------------------------------
    #   Mark translations needing updating
    # -----------------------------------------

    # Check that the golden rule hasn't been broken
    `cmp -s $hotsourcepath $sourcepath`;
    if ($? != 0) {
	print LOG "Error: Source file has changed since backup was made. Please try again.\n";
	die "\n";
    }

    # If the file is not in CVS, we're done
    return if ($option_not_in_cvs == 1);

    # Use CVS to mark each line of the source file with the date it was last edited
    local @sourcedatelist = &cvs_annotate_file($gsdldir, $sourcefile);

    # Build a mapping from key to date
    local %sourcedatehash = &build_key_to_date_mapping(\%sourcelinehash, @sourcedatelist);

    # Check that the golden rule hasn't been broken
    `cmp -s $hottargetpath $targetpath`;
    if ($? != 0) {
	print LOG "Error: Target file has changed since backup was made. Please try again.\n";
	die "\n";
    }

    # Use CVS to mark each line of the target file with the date it was last edited
    local @targetdatelist = &cvs_annotate_file($gsdldir, $targetfile);

    # Build a mapping from key to date
    local %targetdatehash = &build_key_to_date_mapping(\%targetlinehash, @targetdatelist);

    tie(%translationdata, "GDBM_File", $translationdb, GDBM_WRCREAT, 0666);

    # Chunks needing updating are those in the target file that have been more
    #   recently edited in the source file
    foreach $key (keys(%sourcedatehash)) {
	local $sourcedate = $sourcedatehash{$key};
	local $targetdate = $targetdatehash{$key};

	if (defined($targetdate) && &is_date_after($sourcedate, $targetdate)) {
	    # Mark the chunk as needing updating
	    print LOG "Chunk with key $key needs updating.\n";
	    $translationdata{$key . "*"} = "";
	}
    }

    untie(%translationdata);
}


sub get_usage_description
{
    return "Usage: inittranslation.pl <sourceabbr> <targetabbr> <transfile>\n";
}


sub cvs_annotate_file
{
    local ($basedir, $filename) = @_;

    local @dates = ();

    # Use CVS to annotate each line of the file with the date it was last edited
    local $annotatedfile = `cd $basedir; cvs -d $cvsroot annotate -F $filename`;
    local @lines = split(/\n/, $annotatedfile);
    for ($i = 0; $i < scalar(@lines); $i++) {
	local $line = $lines[$i];

	# Extract the date from the CVS annotation at the front
	$line =~ /^\S+\s+\(\S+\s+(\S+)\):\s/;
	local $date = $1;
	push(@dates, $date);
    }

    return @dates;
}


sub build_key_to_date_mapping
{
    local ($linehash, @dates) = @_;

    local %keytodatemap = ();

    foreach $key (keys(%$linehash)) {
	# Chunk lines in linehash start at 0
	local $chunkstart = (split(/-/, $linehash->{$key}))[0];
	local $chunkfinish = (split(/-/, $linehash->{$key}))[1];

	# Get the date the first line of the chunk was last edited
	local $chunkdate = $dates[$chunkstart];

	# Check if any of the other lines in the chunk were edited later
	for ($l = ($chunkstart + 1); $l <= $chunkfinish; $l++) {
            local $linedate = $dates[$l];

	    if (&is_date_after($linedate, $chunkdate)) {
		# This part of the chunk has been updated more recently
		$chunkdate = $linedate;
	    }
	}

	# Map from key to date
	$keytodatemap{$key} = $chunkdate;
    }

    return %keytodatemap;
}


# Returns 1 if $date1 is after $date2, 0 otherwise
sub is_date_after
{
    local ($date1, $date2) = @_;
    local %months = ("Jan", 1, "Feb", 2, "Mar", 3, "Apr",  4, "May",  5, "Jun",  6,
		     "Jul", 7, "Aug", 8, "Sep", 9, "Oct", 10, "Nov", 11, "Dec", 12);

    @date1parts = split(/-/, $date1);
    @date2parts = split(/-/, $date2);

    # Compare year - nasty because we have rolled over into a new century
    $year1 = $date1parts[2];
    if ($year1 < 80) {
	$year1 += 100;
    }
    $year2 = $date2parts[2];
    if ($year2 < 80) {
	$year2 += 100;
    }

    # Compare year
    if ($year1 > $year2) {
	return 1;
    }
    elsif ($year1 == $year2) {
	# Year is the same, so compare month
	if ($months{$date1parts[1]} > $months{$date2parts[1]}) {
	    return 1;
	}
	elsif ($months{$date1parts[1]} == $months{$date2parts[1]}) {
	    # Month is the same, so compare day
	    if ($date1parts[0] > $date2parts[0]) {
		return 1;
	    }
	}
    }

    return 0;
}


&main(@ARGV);
